TEXT_LIST_URL = "https://raw.githubusercontent.com/OpenPecha-dev/editable-text/main/t_text_list.json"
KANGYUR_ARCHIVE_ID = "870326402308432ba173ee3d9043224a"
TENGYUR_ARCHIVE_ID = "187ed94f85154ea5b1ac374a651e1770"
